#include "stacktrace.h"

// Predefined macros in GCC
#ifdef __linux__
 #define LIN_STACK_TRACE
#endif

// Predefined macros in Open Watcom compiler
#ifdef __LINUX__
 #define LIN_STACK_TRACE
#endif

// Predefined macros in Visual C++, Borland C++ and Open Watcom compilers
#ifdef _WIN32
 #define WIN_STACK_TRACE
#endif

// Predefined macros in Watcom compiler
#ifdef __WINDOWS__
 #define WIN_STACK_TRACE
#endif

/**
  The function prints stack trace (i.e. call tree) to the text file defined by out.

  @param out - pointer to the opened file
  @param level - level id from which the stack trace will be printed to out

  @retval The function does not return anything.
*/
void stack_trace(FILE *out, long level) 
{
 #ifdef LIN_STACK_TRACE
  lin_stack_trace(out, level);
 #endif
 #ifdef WIN_STACK_TRACE
  win_stack_trace(out, level);
 #endif
  
}



#ifdef LIN_STACK_TRACE
 #ifndef _GNU_SOURCE
  # define _GNU_SOURCE
 #endif

 #include <stdlib.h>
 #include <dlfcn.h>
 #include <execinfo.h>
 #include <memory.h>
 #include <ucontext.h>
 #ifndef NO_CPP_DEMANGLE
  #include <cxxabi.h>
 #endif


 #if defined(REG_RIP)
  # define STACK_IA64
 #elif defined(REG_EIP)
  # define STACK_X86
 #else
  # define STACK_GENERIC
 #endif

void lin_stack_trace(FILE *out, long level) 
{
  ucontext_t *ucontext = new ucontext_t;
  getcontext(ucontext);
  long count = 0;

 #if defined(STACK_X86) || defined(STACK_IA64)
  Dl_info dlinfo;
  void **bp = 0;
  void *ip = 0;
 #else
  long i;
  void *bt[20];
  char **strings;
  size_t sz;
 #endif

 #if defined(STACK_X86) || defined(STACK_IA64)
  # if defined(STACK_IA64) // 64 bit system
   ip = (void *)ucontext->uc_mcontext.gregs[REG_RIP];
   bp = (void **)ucontext->uc_mcontext.gregs[REG_RBP];
  # elif defined(STACK_X86) // 32 bit system
   ip = (void *)ucontext->uc_mcontext.gregs[REG_EIP];
   bp = (void **)ucontext->uc_mcontext.gregs[REG_EBP];
  # endif

  while(bp && ip) 
  {
    if(!dladdr(ip, &dlinfo))
      break;

    const char *symname = dlinfo.dli_sname;
   #ifndef NO_CPP_DEMANGLE
    int status;
    char *tmp = abi::__cxa_demangle(symname, NULL, NULL, &status);
    if(status == 0 && tmp)
      symname = tmp;
   #endif

    if (count >= level)
    {
      if (count==level)
        fprintf(out, "%s (address %p, raddress %#lx)\n", symname, ip, 
                (unsigned long)(ip) - (unsigned long)(dlinfo.dli_saddr));
      else
        fprintf(out, "  called by %s (address %p raddress %#lx)\n", symname, ip, 
                (unsigned long)(ip) - (unsigned long)(dlinfo.dli_saddr));
    }
   #ifndef NO_CPP_DEMANGLE
    if(tmp)
      free(tmp);
   #endif

    if(dlinfo.dli_sname && !strcmp(dlinfo.dli_sname, "main"))
    {
      fprintf(out, "  in program %s\n", dlinfo.dli_fname);
      break;
    }
    count++;
    ip = (void *)bp[1];
    bp = (void **)bp[0];
  }
 
 #else
  sz = backtrace(bt, 20);
  strings = backtrace_symbols(bt, sz);

  for(i = level; i < sz; ++i)
    fprintf(stderr, "  called by %s()\n", strings[i]);

 #endif
  delete ucontext;
}
#endif



#ifdef  WIN_STACK_TRACE
#include <windows.h>
 #ifdef __WATCOMC__
  #include <imagehlp.h>
 #endif
 #ifndef __WATCOMC__
  #include <dbghelp.h>
 #endif
void win_stack_trace(FILE *out, long level)
{  
  HANDLE hprocess = ::GetCurrentProcess ();
  ::SymInitialize(hprocess, NULL, TRUE);
  CONTEXT context = { 0 };
  HANDLE hthread = ::GetCurrentThread ();
  context.ContextFlags = CONTEXT_FULL;
  ::GetThreadContext (hthread, &context);

  HANDLE hthread2 = NULL;
  DuplicateHandle (hprocess, hthread, hprocess, &hthread2, THREAD_ALL_ACCESS | THREAD_GET_CONTEXT, FALSE, 0);
  context.ContextFlags = CONTEXT_FULL;
  ::SuspendThread(hthread2);
  ::GetThreadContext (hthread2, &context);
  ::ResumeThread(hthread2);
  ::CloseHandle (hthread2);


  // Initialize the STACKFRAME structure for the first call. This is only
  // necessary for Intel CPUs, and isn't mentioned in the documentation.
  STACKFRAME sf = { 0 };
  sf.AddrPC.Offset = context.Eip;
  sf.AddrPC.Mode = AddrModeFlat;
  sf.AddrStack.Offset = context.Esp;
  sf.AddrStack.Mode = AddrModeFlat;
  sf.AddrFrame.Offset = context.Ebp;
  sf.AddrFrame.Mode = AddrModeFlat;


  // Walk the stack up to DATA_LEVELS calls
  int count=0;
  // Get a stack trace for this frame
  while (::StackWalk (IMAGE_FILE_MACHINE_I386, GetCurrentProcess (), GetCurrentThread (), 
                       &sf, &context, 0, SymFunctionTableAccess, SymGetModuleBase, 0)) {

    // Basic sanity check to make sure
    if (sf.AddrFrame.Offset == 0)
      break;

    // IMAGEHLP is wacky, and requires you to pass in a pointer to an
    // IMAGEHLP_SYMBOL structure. The problem is that this structure is
    // variable length. That is, you determine how big the structure is
    // at runtime. This means that you can't use sizeof(struct).
    // So...make a buffer that's big enough, and make a pointer
    // to the buffer. We also need to initialize not one, but TWO
    // members of the structure before it can be used.

    BYTE symbol_buffer[sizeof(IMAGEHLP_SYMBOL) + 512];
    PIMAGEHLP_SYMBOL psymbol = (PIMAGEHLP_SYMBOL)symbol_buffer;
    ::memset (symbol_buffer, 0, sizeof (symbol_buffer));
    psymbol->SizeOfStruct = sizeof(symbol_buffer);
    psymbol->MaxNameLength = 512;

    DWORD sym_displacement = 0;
    ::SymGetSymFromAddr(::GetCurrentProcess (), sf.AddrPC.Offset, &sym_displacement, psymbol);
    // start log from the required stack frame level 
    if (count >= level) {
      if (count==level)
        fprintf(out, "%s (address %p, raddress %#lx)\n", psymbol->Name, context.Eip, 
                (unsigned long)(context.Eip) - (unsigned long)(psymbol->Address));
        fprintf(out, "  called by %s (address %p raddress %#lx)\n", psymbol->Name, context.Eip, 
                (unsigned long)(context.Eip) - (unsigned long)(psymbol->Address));
    }
    count++;
  }
}
#endif
